#include "Shared/RobotInfo.h"
#if (defined(TGT_HAS_ARMS) || !defined(STRICT_TGT_MODEL)) && !defined(TGT_IS_AIBO)

#include "Crew/Grasper.h"
#include "DualCoding/VRmixin.h"
#include "Motion/IKSolver.h"

using namespace DualCoding;

//**** These constants are for the Chiara chess gripper.
float openLeftGripperVal = -0.10f;
float closedLeftGripperVal = -0.25f;
float openRightGripperVal = 0.07f;
float closedRightGripperVal = 0.25f;

GrasperRequest* Grasper::curReq = NULL;
Grasper::GrasperVerbosity_t Grasper::verbosity = -1U;

void Grasper::PlanBodyMotion::doStart() {
  if (curReq->allowBodyMotion) {
    // Ben's code will determine where the robot should be.
    // The code below is just a placeholder for testing purposes..
    NEW_SHAPE(reachPose, AgentData, new AgentData(VRmixin::worldShS, Point(200,0,0,allocentric)));
    reachPose->setColor(rgb(0,255,0));
    reachPose->setObstacle(false);
    curReq->reachPose = reachPose;
  }
    postStateCompletion();
}

void Grasper::FindObj::doStart() {
  if ( curReq->mapreq == NULL ) {
    mapreq = MapBuilderRequest(MapBuilderRequest::worldMap);
    mapreq.addAttributes(curReq->object);
    mapreq.worldTargets.push(new LookoutSearchRequest(curReq->object));
  }
  else
    mapreq = *curReq->mapreq;
}

void Grasper::PlanReachArm::doStart() {
#if defined(TGT_IS_CALLIOPE2)
  NodeValue_t armpoint;
  armpoint[0] = 0.75;  // arm base
  armpoint[1] = 0; // arm shoulder
  curReq->reachPath.push_back(armpoint);
  armpoint[1] = -0.60; // arm shoulder
  curReq->reachPath.push_back(armpoint);
  postStateCompletion();
#elif defined(TGT_IS_CALLIOPE5)
  // plan Calliope5 arm reach
  // assume there is a shape in local space that matches our world space object; fail if not
#elif defined(TGT_HANDEYE)
  std::vector<NodeValue_t> endStates;
  Point toPt = curReq->object->getCentroid();
  VRmixin::grasper->computeGoalStates(toPt, curReq->gripperAngleRangesX, curReq->gripperAngleRangesY, curReq->gripperAngleRangesZ,
                                      curReq->angleResolution, endStates, fmat::pack(0,0,0));
  if ( endStates.empty() ) {
    postStateSignal<GraspError>(GrasperRequest::pickUpUnreachable);
    return;
  }

  NodeValue_t startSt;
  Grasper::getCurrentState(startSt);
  
  VRmixin::grasper->appendObjectToObstacles();
  
  GET_SHAPE(worldBounds, PolygonData, VRmixin::worldShS);
  Grasper::Planner planner(VRmixin::worldShS, worldBounds, curReq->rrtInflation, curReq->effectorOffset);
  std::vector<NodeType_t> *treeStartResult = curReq->displayTree ? new std::vector<NodeType_t> : NULL;
  std::vector<NodeType_t> *treeEndResult = curReq->displayTree ? new std::vector<NodeType_t> : NULL;
  fmat::Transform t;
  t.translation() = VRmixin::theAgent->getCentroid().coords;
  t.rotation() = fmat::rotationZ(VRmixin::theAgent->getOrientation());
  std::cout << "Begin plan." << std::endl;
  Grasper::Planner::PlannerResult result =
    planner.planPath(startSt, endStates.front(), curReq->rrtInterpolationStep,
		     t, curReq->rrtMaxIterations,
		     &(curReq->reachPath), treeStartResult, treeEndResult);
  VRmixin::grasper->removeTargetFromObstacles();
  std::cout << "End plan." << std::endl;
  
  switch ( result.code ) {
    case GenericRRTBase::SUCCESS:
      std::cout << "Plan Unconstrained succeeded" << std::endl;
      break;
    case GenericRRTBase::START_COLLIDES:
      std::cout << "Arm path planning failed: start state is in collision.\n";
      break;
    case GenericRRTBase::END_COLLIDES:
      std::cout << "Arm path planning failed: end state is in collision.\n";
      break;
    case GenericRRTBase::MAX_ITER:
      std::cout << "Arm path planning failed: too many iterations.\n";
      break;
  }
  
  if (result.code == GenericRRTBase::START_COLLIDES || result.code == GenericRRTBase::END_COLLIDES) {
    std::cout << "Obstacle: " << std::endl << "=================" << std::endl
	      << result.movingObstacle->toString() << std::endl;
    std::cout << "collided with: " << std::endl << "=================" << std::endl
	      << result.collidingObstacle->toString() << std::endl;
    std::cout << "=================" << std::endl;
  }
  
  // Display the tree if requested
  if ( curReq->displayTree && result.code != GenericRRTBase::START_COLLIDES && GenericRRTBase::END_COLLIDES ) {
    NEW_SHAPE(plannerTrees, GraphicsData, new GraphicsData(VRmixin::worldShS));
    planner.plotTree(*treeStartResult, plannerTrees, rgb(0,0,0));
    planner.plotTree(*treeEndResult, plannerTrees, rgb(0,192,0));
    delete treeStartResult;
    delete treeEndResult;
  }
  
  // Display the path if requested
  if ( curReq->displayPath && result.code == GenericRRTBase::SUCCESS ) {
    NEW_SHAPE(plannedPath, GraphicsData, new GraphicsData(VRmixin::worldShS));
    planner.plotPath(curReq->reachPath, plannedPath, rgb(0,0,255));
  } 
  
  if (result.code != GenericRRTBase::SUCCESS) {
    std::cout << "PLANNER FAILURE - ARM REACH" << std::endl;
    postStateSignal<GraspError>(GrasperRequest::noGraspPath);
    return;
  }
  postStateCompletion();
#endif
}

void Grasper::GraspBody::doStart() {
#ifdef TGT_IS_CALLIOPE2
  ShapeRoot objlocal = find_if(localShS, IsLastMatch(curReq->object));
  if ( ! objlocal.isValid() ) {
    std::cout << "Can't find " << curReq->object << " last match id " << curReq->object->getLastMatchId()
	      << " in local shape space!" << std::endl;
    cancelThisRequest();
    return;
  }
  // Thomas's code goes here
  // Construct a waypoint sequence to grab the object
#else
  cancelThisRequest();
#endif
}

void Grasper::GraspArm::doStart() {
  postStateCompletion();
}


// Comprehensive planning function for grasp behaviors
void Grasper::Plan::doStart() {
  curReq->reachPath.clear();
  curReq->movePath.clear();
  curReq->releasePath.clear();
  /*
  switch (curReq->requestType) {
  case GrasperRequest::grasp:
  case GrasperRequest::checkGraspable:
  case GrasperRequest::reach:
  case GrasperRequest::computeReach:
  case GrasperRequest::touch:
  case GrasperRequest::sweep:   // should really have its own planner class
  case GrasperRequest::moveTo:
  case GrasperRequest::computeMove:
  case GrasperRequest::checkMovable:
    result = planReachBody();
    if ( result != GrasperRequest::noError ) {
      postStateSignal<GraspError>(result);
      return;
    }
    result = planReachArm();
    if ( result != GrasperRequest::noError ) {
      postStateSignal<GraspError>(result);
      return;
    }
    break;
  default:
    break;
  }

  switch (curReq->requestType) {
  case GrasperRequest::moveTo:
  case GrasperRequest::computeMove:
  case GrasperRequest::checkMovable:
    result = planMoveBody();
    if ( result != GrasperRequest::noError ) {
      postStateSignal<GraspError>(result);
      return;
    }
    result = planMoveArm();
    if ( result != GrasperRequest::noError ) {
      postStateSignal<GraspError>(result);
      return;
    }
    break;
  default:
    break;
  }

  switch (curReq->requestType) {
  case GrasperRequest::moveTo:
  case GrasperRequest::computeMove:
  case GrasperRequest::checkMovable:
  case GrasperRequest::release:
  case GrasperRequest::rest:
  case GrasperRequest::computeRest:
  case GrasperRequest::checkRestable:
    result = planRestArm();
    if ( result != GrasperRequest::noError ) {
      postStateSignal<GraspError>(result);
      return;
    }
    result = planRestBody();
    if ( result != GrasperRequest::noError ) {
      postStateSignal<GraspError>(result);
      return;
    }
    break;
  default:
    break;
  }
  */
  postStateCompletion();
}

void Grasper::PathPlanConstrained::doStart() {
  std::vector<NodeValue_t> endStates;
  
  if (!curReq->targetLocation.isValid()) {
    postStateSignal<GraspError>(GrasperRequest::someError);
    return;
  }
  
  Point toPt = curReq->object->getCentroid();
  VRmixin::grasper->computeGoalStates(toPt, curReq->gripperAngleRangesX, curReq->gripperAngleRangesY, curReq->gripperAngleRangesZ,
                                      curReq->angleResolution, endStates, fmat::pack(0,0,0));
  if (endStates.empty()) {
    std::cout << "PLAN - pickUpUnreachable" << std::endl;
    postStateSignal<GraspError>(GrasperRequest::pickUpUnreachable);
    return;
  }
  
  NodeValue_t startSt;
  if (curReq->reachPath.empty())
    Grasper::getCurrentState(startSt);
  else
    startSt = curReq->reachPath.back();
  
  GET_SHAPE(worldBounds, PolygonData, VRmixin::worldShS);
  Planner planner(VRmixin::worldShS, worldBounds, curReq->rrtInflation,
                                             curReq->effectorOffset, curReq->predicate);
  
  std::vector<NodeType_t> *treeStartResult = NULL;
  std::vector<NodeType_t> *treeEndResult = NULL;
  fmat::Transform t;
  t.translation() = VRmixin::theAgent->getCentroid().coords;
  t.rotation() = fmat::rotationZ(VRmixin::theAgent->getOrientation());
	Planner::PlannerResult result =
	planner.planPath(startSt, endStates.front(), curReq->rrtInterpolationStep,
									 t, curReq->rrtMaxIterations,
									 &(curReq->reachPath), treeStartResult, treeEndResult);
  
  switch ( result.code ) {
    case GenericRRTBase::SUCCESS:
      std::cout << "Plan Constrained succeeded" << std::endl;
      postStateCompletion();
      return;
    case GenericRRTBase::START_COLLIDES:
      std::cout << "Navigation path planning failed: start state is in collision.\n";
      break;
    case GenericRRTBase::END_COLLIDES:
      std::cout << "Navigation path planning failed: end state is in collision.\n";
      break;
    case GenericRRTBase::MAX_ITER:
      std::cout << "Navigation path planning failed: too many iterations.\n";
      break;
  }
  
  if (result.code == GenericRRTBase::START_COLLIDES || result.code == GenericRRTBase::END_COLLIDES) {
    std::cout << "Obstacle: " << std::endl << "=================" << std::endl << result.movingObstacle->toString() << std::endl;
    std::cout << "collided with: " <<std::endl << "=================" << std::endl << result.collidingObstacle->toString() << std::endl;
    std::cout << "=================" << std::endl;
  }
  
  // Display the tree if requested
  if ( curReq->displayTree && result.code != GenericRRTBase::START_COLLIDES && GenericRRTBase::END_COLLIDES ) {
    NEW_SHAPE(plannerTrees, GraphicsData, new GraphicsData(VRmixin::worldShS));
    planner.plotTree(*treeStartResult, plannerTrees, rgb(0,0,0));
    planner.plotTree(*treeEndResult, plannerTrees, rgb(0,192,0));
    delete treeStartResult;
    delete treeEndResult;
  }
  
  // Display the path if requested
  if ( curReq->displayPath && result.code == GenericRRTBase::SUCCESS ) {
    NEW_SHAPE(plannedPath, GraphicsData, new GraphicsData(VRmixin::worldShS));
    planner.plotPath(curReq->movePath, plannedPath, rgb(0,0,255));
  } 
  
  if (result.code == GenericRRTBase::SUCCESS)
    postStateCompletion();
  else {
    std::cout << "PLANNER FAILURE - CONSTRAINED" << std::endl;
    postStateSignal<GraspError>(GrasperRequest::noGraspPath);
  }
}

void Grasper::PathPlanToRest::doStart() {
  NodeValue_t startSt;
  NodeValue_t endSt;
  
  switch(curReq->restType) {
    case GrasperRequest::stationary:
      postStateCompletion();
      return;
    case GrasperRequest::settleArm:
    case GrasperRequest::settleBodyAndArm:
      for(unsigned i = 0; i < NumArmJoints; i++)
        endSt[i] = curReq->armRestState[i];
      break;
  }
  
  VRmixin::grasper->appendObjectToObstacles();
  
  if (!curReq->movePath.empty())
    startSt = curReq->movePath.back();
  else if (!curReq->reachPath.empty())
    startSt = curReq->reachPath.back();
  else
    Grasper::getCurrentState(startSt);
  
  GET_SHAPE(worldBounds, PolygonData, VRmixin::worldShS);
  Planner planner(VRmixin::worldShS, worldBounds, curReq->rrtInflation, curReq->effectorOffset, NULL);
  
  std::vector<NodeType_t> *treeStartResult = NULL;
  std::vector<NodeType_t> *treeEndResult = NULL;
  fmat::Transform t;
  t.translation() = VRmixin::theAgent->getCentroid().coords;
  t.rotation() = fmat::rotationZ(VRmixin::theAgent->getOrientation());
  Planner::PlannerResult result =
    planner.planPath(startSt, endSt, curReq->rrtInterpolationStep,
		     t, curReq->rrtMaxIterations,
		     &(curReq->releasePath), treeStartResult, treeEndResult);
  VRmixin::grasper->removeTargetFromObstacles();
  
  switch ( result.code ) {
    case GenericRRTBase::SUCCESS:
      std::cout << "Plan disengage succeeded" << std::endl;
      postStateCompletion();
      return;
    case GenericRRTBase::START_COLLIDES:
      std::cout << "Navigation path planning failed: start state is in collision.\n";
      break;
    case GenericRRTBase::END_COLLIDES:
      std::cout << "Navigation path planning failed: end state is in collision.\n";
      break;
    case GenericRRTBase::MAX_ITER:
      std::cout << "Navigation path planning failed: too many iterations.\n";
      break;
  }
  
  if (result.code == GenericRRTBase::START_COLLIDES || result.code == GenericRRTBase::END_COLLIDES) {
    std::cout << "Obstacle: " << std::endl << "=================" << std::endl << result.movingObstacle->toString() << std::endl;
    std::cout << "collided with: " <<std::endl << "=================" << std::endl << result.collidingObstacle->toString() << std::endl;
    std::cout << "=================" << std::endl;
  }
  
  // Display the tree if requested
  if ( curReq->displayTree && result.code != GenericRRTBase::START_COLLIDES && GenericRRTBase::END_COLLIDES ) {
    NEW_SHAPE(plannerTrees, GraphicsData, new GraphicsData(VRmixin::worldShS));
    planner.plotTree(*treeStartResult, plannerTrees, rgb(0,0,0));
    planner.plotTree(*treeEndResult, plannerTrees, rgb(0,192,0));
    delete treeStartResult;
    delete treeEndResult;
  }
  
  // Display the path if requested
  if ( curReq->displayPath && result.code == GenericRRTBase::SUCCESS ) {
    NEW_SHAPE(plannedPath, GraphicsData, new GraphicsData(VRmixin::worldShS));
    planner.plotPath(curReq->releasePath, plannedPath, rgb(0,0,255));
  } 
  
  if (result.code == GenericRRTBase::SUCCESS)
    postStateCompletion();
  else {
    std::cout << "PLANNER FAILURE - ARM_REST" << std::endl;
    postStateSignal<GraspError>(GrasperRequest::noGraspPath);
  }
}

void Grasper::MoveArm::doStart() {
  switch(pa) {
    case GrasperRequest::doReach:
      if ( curReq->verbosity & GVexecutePath )
        std::cout << "Grasper: executing arm reach path" << std::endl;
      executeMove(curReq->reachPath);
      break;
    case GrasperRequest::doMove:
      if ( curReq->verbosity & GVexecutePath )
        std::cout << "Grasper: moving along constrained path" << std::endl;
      executeMove(curReq->movePath);
      break;
    case GrasperRequest::doRelease:
      if ( curReq->verbosity & GVexecutePath )
        std::cout << "Grasper: executing arm release path" << std::endl;
      executeMove(curReq->releasePath);
      break;
    case GrasperRequest::noPath:
      break;
  }
}

void Grasper::MoveArm::executeMove(const std::vector<NodeValue_t>& path) {
#ifdef TGT_HAS_ARMS
  MMAccessor<DynamicMotionSequence> move_acc = getMC();
  move_acc->clear();
  move_acc->setTime(10);
  for(unsigned int p = 0; p < path.size(); p++) {
    for(unsigned int j = 0; j < numPlannerJoints; j++) {
      if (kine->getKinematicJoint(ArmOffset+j)->isMobile() == false)
        continue;
      move_acc->setOutputCmd(ArmOffset+j, (float)path[p][j]);
    }
    if (p < path.size()-1) {
      NodeValue_t confDif;
      for(unsigned i = 0; i < numPlannerJoints; i++)
        confDif[i] = path[p][i] - path[p+1][i];
      move_acc->advanceTime( advTime(confDif) );
    }
  }
  
  move_acc->play();
#endif
}

void Grasper::SetJoint::moveJoint(float value) {
  unsigned int offset = 10;
  if(jointName == "ARM:shldr")
    offset = 0;
  else if(jointName == "ARM:elbow")
    offset = 1;
  else if(jointName == "ARM:wristYaw")
    offset = 2;
  else if(RobotInfo::NumArmJoints > 3) {
    if(jointName == "ARM:wristPitch")
      offset = 3;
    else if(jointName == "ARM:wristRoll" || jointName == "ARM:gripperLeft")
      offset = 4;
    else if(jointName == "ARM:gripper" || jointName == "ARM:gripperRight")
      offset = 5;
    else {
      postStateCompletion();
      return;
    }
  }
  else {
    postStateCompletion();
    return;
  }
  if ( (curReq ? curReq->verbosity : verbosity) & GVsetJoint )
    cout << "Setting " << jointName << " to " << value << endl;
  getMC()->takeSnapshot();
  getMC()->setMaxSpeed(offset, speed); 	 
  getMC()->setJointValue(offset, value);
} 

void Grasper::getCurrentState(NodeValue_t &current, KinematicJoint* endEffector) {
  KinematicJoint* joint = !endEffector ? kine->getKinematicJoint(curReq->effectorOffset)->cloneBranch() : endEffector;
  for (unsigned int j = numPlannerJoints; j > 0; j--) {
    while (joint->isMobile() == false) joint = joint->getParent();
    current[j-1] = joint->getQ();
    joint = joint->getParent();
  }
}

// TODO: why is this here?
void Grasper::computeGoalStates(Point& toPt, std::vector<std::pair<float, float> >& rangesX,
                                std::vector<std::pair<float, float> >& rangesY, std::vector<std::pair<float, float> >& rangesZ,
                                float res, std::vector<NodeValue_t>& goals, fmat::Column<3> offset) {
  KinematicJoint* effector = kine->getKinematicJoint(curReq->effectorOffset)->cloneBranch();
  IKSolver& solver = effector->getIK();
  goals.clear();
  if (res == 0)
    res = 1;
  
  if (rangesX.empty()) {
    rangesX.push_back(pair<float, float>(0, 2*M_PI));
  }
  
  if (rangesY.empty()) {
    rangesY.push_back(pair<float, float>(0, 2*M_PI));
  }
  
  if (rangesZ.empty()) {
    rangesZ.push_back(pair<float, float>(0, 2*M_PI));
  }
  
  int factorX = 0;
  bool keepGoing = true;
  do {
    keepGoing = false;
    for(unsigned rX = 0; rX < rangesX.size(); rX++) {
      
      float thetaX = res * factorX;
      float midX = (rangesX[rX].first + rangesX[rX].second)/2;
      if (thetaX > rangesX[rX].second - midX)
        continue;
      int factorY = 0;
      
      for (unsigned rY = 0; rY < rangesY.size(); rY++) {
        
        float thetaY = res * factorY;
        float midY = (rangesY[rY].first + rangesY[rY].second)/2;
        if (thetaY > rangesY[rY].second - midY)
          continue;
        int factorZ = 0;
        
        for (unsigned rZ = 0; rZ < rangesZ.size(); rZ++) {
          
          float thetaZ = res * factorZ;
          float midZ = (rangesZ[rZ].first + rangesZ[rZ].second)/2;
          if (thetaZ > rangesZ[rZ].second - midZ)
            continue;
          
          keepGoing = true;
          fmat::Quaternion ori = fmat::Quaternion::fromMatrix(fmat::rotationZ(midZ + thetaZ) * fmat::rotationY(midY + thetaY) * fmat::rotationX(midX + thetaX));
          bool reached;
          reached = solver.solve(offset, IKSolver::Rotation(), *effector, IKSolver::Point(toPt.coords[0], toPt.coords[1], toPt.coords[2]), 1, IKSolver::Rotation(ori), 0);
          if (reached) {
            NodeValue_t endSt;
            KinematicJoint* joint = effector;
            for (unsigned int j = numPlannerJoints; j > 0; j--) {
              while (joint->isMobile() == false) joint = joint->getParent();
              endSt[j-1] = joint->getQ();
              joint = joint->getParent();
            }
            goals.push_back(endSt);
          }
          
          ori = fmat::Quaternion::fromMatrix(fmat::rotationZ(midZ - thetaZ) * fmat::rotationY(midY - thetaY) * fmat::rotationX(midX - thetaX));
          reached = solver.solve(offset, IKSolver::Rotation(), *effector, IKSolver::Point(toPt.coords[0], toPt.coords[1], toPt.coords[2]), 1, IKSolver::Rotation(ori), 0);
          if (reached) {
            NodeValue_t endSt;
            KinematicJoint* joint = effector;
            for(unsigned int j = numPlannerJoints; j > 0; j--) {
              while (joint->isMobile() == false) joint = joint->getParent();
              endSt[j-1] = joint->getQ();
              joint = joint->getParent();
            }
            goals.push_back(endSt);
          }
          if ( curReq->maxNumberOfAngles <= (signed)goals.size() ) {
            keepGoing = false;
          }
          factorZ++;
        } // Z
        factorY++;
      } // Y
      factorX++;
    } // X
  } while (keepGoing);
  if ( curReq->verbosity & GVcomputeGoals )
    std::cout << "Grasper found " << goals.size() << " potential goal states." << std::endl;
}

unsigned int Grasper::executeRequest(const GrasperRequest& req, BehaviorBase* requestingBehavior) {
  GrasperRequest *newReq = new GrasperRequest(req);
  newReq->requestingBehavior = requestingBehavior;
  unsigned int reqID = ++idCounter;
  newReq->requestID = reqID;
  GraspError e = newReq->validateRequest();
  if ( e != GrasperRequest::noError ) {
    std::cout << "*** Grasper received invalid grasp request, id=" << reqID << std::endl;
    GrasperEvent myEvent(false, newReq->requestType, e, (size_t)newReq->requestingBehavior);
    erouter->postEvent(myEvent);
  }
  else {
    requests.push(newReq);
    executeRequest();
  }
  return reqID;
}

void Grasper::executeRequest() {
  if ( curReq != NULL )
    return;
  else if ( requests.empty() )
    return;
  curReq = requests.front();
  requests.pop();
  
  curReq->verbosity = (verbosity & ~curReq->clearVerbosity) | curReq->setVerbosity;
  if ( curReq->verbosity & GVexecuteRequest )
    std::cout << "Grasper now executing request " << curReq->requestID << std::endl;
  
  /*
  if (curReq->object.isValid() && curReq->object->getSpace().getRefFrameType() == VRmixin::worldShS.getRefFrameType())
    curReq->object = VRmixin::mapBuilder->importWorldToLocal(curReq->object);
  
  if (curReq->targetLocation.isValid() && curReq->targetLocation->getSpace().getRefFrameType() == VRmixin::worldShS.getRefFrameType())
    curReq->targetLocation = VRmixin::mapBuilder->importWorldToLocal(curReq->targetLocation);
  */
  {
    GET_SHAPE(reachPose, AgentData, VRmixin::worldShS);
    GET_SHAPE(releasePose, AgentData, VRmixin::worldShS);
    GET_SHAPE(restPose, AgentData, VRmixin::worldShS);
    GET_SHAPE(plannerTrees, GraphicsData, VRmixin::worldShS);
    if ( reachPose.isValid() )
      VRmixin::worldShS.deleteShape(reachPose);
    if ( releasePose.isValid() )
      VRmixin::worldShS.deleteShape(releasePose);
    if ( restPose.isValid() )
      VRmixin::worldShS.deleteShape(restPose);
    if ( plannerTrees.isValid() )
      VRmixin::worldShS.deleteShape(plannerTrees);
  }

  dispatch();
}

void Grasper::dispatch() {
  startmain_->start();
  /*
  switch( curReq->requestType ) {
    case GrasperRequest::reach:
      reachRoot->start();
      break;
    case GrasperRequest::computeReach:
      if ( curReq->populateEventPathWith == GrasperRequest::noPath )
        curReq->populateEventPathWith = GrasperRequest::moveFree;
      computeReachRoot->start();
      break;
    case GrasperRequest::grasp:
      graspRoot->start();
      break;
    case GrasperRequest::checkGraspable:
      checkRoot->start();
      break;
    case GrasperRequest::release:
      releaseRoot->start();
      break;
    case GrasperRequest::touch:
      touchRoot->start();
      break;
    case GrasperRequest::moveTo:
      moveRoot->start();
      break;
    case GrasperRequest::checkMovable:
      checkRoot->start();
      break;
    case GrasperRequest::rest:
      restRoot->start();
      break;
    case GrasperRequest::checkRestable:
      checkRoot->start();
      break;
    case GrasperRequest::sweep:
      std::cout << "Grasper functionality " << curReq->requestType << " not implemented"<< std::endl;
      break;
    default:
      std::cout << "Unidentified requestType" << std::endl;
      break;
  }
  */
} 

#endif // TGT_HAS_ARMS
